#!/usr/bin/env python
#
# __init__.py - ARB vertex/fragment program management.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``arbp`` package is the home of the :class:`.ARBPShader` class, a class
which may be used to manage OpenGL ``ARB_vertex_program`` and
``ARB_fragment_program`` shader programs.  The :class:`.ARBPShader` class is
defined in the :mod:`.program` module, and uses some functions defined in the
:mod:`.parse` module.

See also the :mod:`.glsl` package.
"""
